# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Dict
from typing import List
from typing import Any
from typing import Optional

from ooresults.pdf.pdf import PDF
from ooresults.repo.series_type import Settings


def create_pdf(
    settings: Settings, events: List, results: List, landscape: bool = False
) -> bytes:
    W_SPACE = 3
    W_RANK = 9
    W_NAME = 40
    W_YEAR = 10
    W_SUM = 15 + 2
    W_POINTS = 15 + 2

    # to set the results right aligned with the right margin (minus W_RANK) we extend the club space
    W_CLUB = 297 - 20 if landscape else 210 - 20
    W_CLUB -= (W_RANK + W_SPACE + W_NAME + W_SPACE + W_YEAR + W_SPACE) + W_RANK
    W_CLUB -= W_SUM + len(events) * W_POINTS

    # Instantiation of inherited class
    pdf = PDF(name=settings.name, landscape=landscape)
    pdf.set_margin(margin=10)
    pdf.set_auto_page_break(auto=True, margin=20)
    pdf.add_page()

    def cell(w: int, h: Optional[int] = None, txt: str = "", align: str = "L") -> None:
        while pdf.get_string_width(txt) > w:
            txt = txt[:-1]
        pdf.cell(w=w, h=h, txt=txt, align=align)

    def format_points(points: float) -> str:
        if points is not None:
            return str(points)
        else:
            return "-----"

    for i, class_results in enumerate(results):
        class_name, ranked_results = class_results
        if i > 0:
            # insert a page break if there is not enough space left on the
            # page for the class header, the table header and two table rows
            if pdf.will_page_break(height=30):
                pdf.add_page()
            else:
                pdf.ln()
                pdf.ln()

        pdf.set_font(family="Carlito", style="B", size=12)
        pdf.cell(txt=class_name)
        pdf.ln()
        pdf.ln()
        pdf.set_font(family="Carlito", style="I", size=10)
        cell(w=W_RANK, h=None, txt="Pl", align="R")
        cell(w=W_SPACE, h=None, txt="")
        cell(w=W_NAME, h=None, txt="Name", align="L")
        cell(w=W_SPACE, h=None, txt="")
        cell(w=W_YEAR, h=None, txt="Jg", align="R")
        cell(w=W_SPACE, h=None, txt="")
        cell(w=W_CLUB, h=None, txt="Verein", align="L")
        cell(w=W_SPACE, h=None, txt="")
        cell(w=W_SUM, h=None, txt="Gesamt", align="R")
        for e in events:
            cell(w=W_POINTS, h=None, txt=e.series, align="R")
        pdf.ln()
        pdf.ln()

        for result in ranked_results:

            def get(d: Dict, key: str) -> Any:
                s = d.get(key, "")
                return s if s is not None else ""

            if result["races"] != {}:
                pdf.set_font(family="Carlito", size=10)
                cell(
                    w=W_RANK,
                    h=None,
                    txt=str(get(result, "rank"))
                    if str(get(result, "rank")) != ""
                    else "-",
                    align="R",
                )
                cell(w=W_SPACE, h=None, txt="")
                cell(
                    w=W_NAME,
                    h=None,
                    txt=get(result, "last_name") + " " + get(result, "first_name"),
                    align="L",
                )
                cell(w=W_SPACE, h=None, txt="")
                cell(w=W_YEAR, h=None, txt=str(get(result, "year")), align="R")
                cell(w=W_SPACE, h=None, txt="")
                cell(w=W_CLUB, h=None, txt=get(result, "club"), align="L")
                cell(w=W_SPACE, h=None, txt="")
                pdf.set_font(style="B")
                cell(
                    w=W_SUM,
                    h=None,
                    txt=format_points(result.get("sum", None)),
                    align="R",
                )
                pdf.set_font()
                for i in range(len(events)):
                    if i in result["organizer"].keys():
                        cell(
                            w=W_POINTS,
                            h=None,
                            txt="("
                            + format_points(result["organizer"].get(i, None))
                            + ")",
                            align="R",
                        )
                    else:
                        cell(
                            w=W_POINTS,
                            h=None,
                            txt=format_points(result["races"].get(i, None)),
                            align="R",
                        )
                pdf.ln()

    return bytes(pdf.output())



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

import pytest

from ooresults.repo import repo
from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.class_params import ClassParams
from ooresults.repo.result_type import ResultStatus


@pytest.fixture
def db():
    return SqliteRepo(db=":memory:")


@pytest.fixture
def event_1_id(db):
    return db.add_event(
        name="Event 1",
        date=datetime.date(year=2020, month=1, day=1),
        key=None,
        publish=False,
        series=None,
    )


@pytest.fixture
def event_2_id(db):
    return db.add_event(
        name="Event 2",
        date=datetime.date(year=2020, month=1, day=1),
        key=None,
        publish=False,
        series=None,
    )


@pytest.fixture
def course_id(db, event_1_id):
    return db.add_course(
        event_id=event_1_id,
        name="Course 1",
        length=None,
        climb=None,
        controls="",
    )


@pytest.fixture
def class_1_id(db, event_1_id):
    return db.add_class(
        event_id=event_1_id,
        name="Class 1",
        short_name="C 1",
        course_id=None,
        params=ClassParams(),
    )


@pytest.fixture
def class_2_id(db, event_1_id, course_id):
    return db.add_class(
        event_id=event_1_id,
        name="Class 2",
        short_name=None,
        course_id=course_id,
        params=ClassParams(),
    )


@pytest.fixture
def class_3_id(db, event_2_id):
    return db.add_class(
        event_id=event_2_id,
        name="Class 3",
        short_name="C 3",
        course_id=None,
        params=ClassParams(),
    )


@pytest.fixture
def entry_id(db, event_1_id, class_1_id):
    return db.add_entry(
        event_id=event_1_id,
        competitor_id=None,
        first_name="A",
        last_name="B",
        gender="",
        year=None,
        class_id=class_1_id,
        club_id=None,
        not_competing=False,
        chip="",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )


def test_get_classes_after_adding_one_class(db, event_1_id, class_1_id):
    c = list(db.get_classes(event_id=event_1_id))
    assert len(c) == 1
    assert c[0].id == class_1_id
    assert c[0].name == "Class 1"
    assert c[0].short_name == "C 1"
    assert c[0].course_id == None
    assert c[0].course == None
    assert c[0].params == ClassParams()


def test_get_classes_for_first_event(
    db, event_1_id, course_id, class_1_id, class_2_id, class_3_id
):
    c = list(db.get_classes(event_id=event_1_id))
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == class_1_id
    assert c[0].name == "Class 1"
    assert c[0].short_name == "C 1"
    assert c[0].course_id == None
    assert c[0].course == None
    assert c[0].params == ClassParams()
    assert c[1].id == class_2_id
    assert c[1].name == "Class 2"
    assert c[1].short_name == None
    assert c[1].course_id == course_id
    assert c[1].course == "Course 1"
    assert c[1].params == ClassParams()


def test_get_classes_for_second_event(
    db, event_2_id, course_id, class_1_id, class_2_id, class_3_id
):
    c = list(db.get_classes(event_id=event_2_id))
    assert len(c) == 1

    assert c[0].id == class_3_id
    assert c[0].name == "Class 3"
    assert c[0].short_name == "C 3"
    assert c[0].course_id == None
    assert c[0].course == None
    assert c[0].params == ClassParams()


def test_get_first_added_class(db, class_1_id, class_2_id):
    c = list(db.get_class(id=class_1_id))
    assert len(c) == 1
    assert c[0].id == class_1_id
    assert c[0].name == "Class 1"
    assert c[0].short_name == "C 1"
    assert c[0].course_id == None
    assert c[0].params == ClassParams()


def test_get_last_added_class(db, course_id, class_1_id, class_2_id):
    c = list(db.get_class(id=class_2_id))
    assert len(c) == 1
    assert c[0].id == class_2_id
    assert c[0].name == "Class 2"
    assert c[0].short_name == None
    assert c[0].course_id == course_id
    assert c[0].params == ClassParams()


def test_update_first_added_class(db, event_1_id, course_id, class_1_id, class_2_id):
    db.update_class(
        id=class_1_id,
        name="Class 3",
        short_name="C 3",
        course_id=course_id,
        params=ClassParams(),
    )
    c = list(db.get_classes(event_id=event_1_id))
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == class_2_id
    assert c[0].name == "Class 2"
    assert c[0].short_name == None
    assert c[0].course_id == course_id
    assert c[0].course == "Course 1"
    assert c[0].params == ClassParams()
    assert c[1].id == class_1_id
    assert c[1].name == "Class 3"
    assert c[1].short_name == "C 3"
    assert c[1].course_id == course_id
    assert c[1].course == "Course 1"
    assert c[1].params == ClassParams()


def test_update_last_added_class(db, event_1_id, class_1_id, class_2_id):
    db.update_class(
        id=class_2_id,
        name="Class 3",
        short_name="C 3",
        course_id=None,
        params=ClassParams(),
    )
    c = list(db.get_classes(event_id=event_1_id))
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == class_1_id
    assert c[0].name == "Class 1"
    assert c[0].short_name == "C 1"
    assert c[0].course_id == None
    assert c[0].course == None
    assert c[0].params == ClassParams()
    assert c[1].id == class_2_id
    assert c[1].name == "Class 3"
    assert c[1].short_name == "C 3"
    assert c[1].course_id == None
    assert c[1].course == None
    assert c[1].params == ClassParams()


def test_delete_first_added_class(db, event_1_id, course_id, class_1_id, class_2_id):
    db.delete_class(id=class_1_id)
    c = list(db.get_classes(event_id=event_1_id))
    assert len(c) == 1
    assert c[0].id == class_2_id
    assert c[0].name == "Class 2"
    assert c[0].short_name == None
    assert c[0].course_id == course_id
    assert c[0].course == "Course 1"
    assert c[0].params == ClassParams()


def test_delete_last_added_class(db, event_1_id, class_1_id, class_2_id):
    db.delete_class(id=class_2_id)
    c = list(db.get_classes(event_id=event_1_id))
    assert len(c) == 1
    assert c[0].id == class_1_id
    assert c[0].name == "Class 1"
    assert c[0].short_name == "C 1"
    assert c[0].course_id == None
    assert c[0].course == None
    assert c[0].params == ClassParams()


def test_delete_classes_deletes_all_classes(db, event_1_id, class_1_id, class_2_id):
    db.delete_classes(event_id=event_1_id)
    c = list(db.get_classes(event_id=event_1_id))
    assert len(c) == 0


def test_add_existing_name_raises_exception(db, event_1_id, class_1_id):
    with pytest.raises(repo.ConstraintError, match="Class already exist"):
        db.add_class(
            event_id=event_1_id,
            name="Class 1",
            short_name="XXX",
            course_id=None,
            params=ClassParams(),
        )


def test_change_to_existing_name_raises_exception(db, class_1_id, class_2_id):
    with pytest.raises(repo.ConstraintError, match="Class already exist"):
        db.update_class(
            id=class_1_id,
            name="Class 2",
            short_name=None,
            course_id=None,
            params=ClassParams(),
        )


def test_update_with_unknown_id_raises_exception(db, class_1_id):
    with pytest.raises(KeyError):
        db.update_class(
            id=class_1_id + 1,
            name="Class 2",
            short_name="C 2",
            course_id=None,
            params=ClassParams(),
        )


def test_add_class_with_unknown_event_id_raises_exception(db, event_1_id):
    with pytest.raises(repo.EventNotFoundError):
        db.add_class(
            event_id=event_1_id + 1,
            name="Class 1",
            short_name="C 1",
            course_id=None,
            params=ClassParams(),
        )


def test_delete_class_with_unknown_id_do_not_change_anything(
    db, event_1_id, class_1_id
):
    db.delete_class(id=class_1_id + 1)
    c = list(db.get_classes(event_id=event_1_id))
    assert len(c) == 1
    assert c[0].id == class_1_id
    assert c[0].name == "Class 1"
    assert c[0].short_name == "C 1"
    assert c[0].course_id == None
    assert c[0].course == None
    assert c[0].params == ClassParams()


def test_delete_class_used_in_entry_raises_exception(
    db, event_1_id, entry_id, class_1_id
):
    with pytest.raises(repo.ClassUsedError):
        db.delete_class(id=class_1_id)


def test_delete_classes_if_entry_exist_raises_exception(
    db, event_1_id, entry_id, class_1_id
):
    with pytest.raises(repo.ClassUsedError):
        db.delete_classes(event_id=event_1_id)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import copy
import json
import pathlib
import datetime
import bz2
import functools
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import Dict

import tzlocal
import websockets
from websockets.legacy.server import WebSocketServerProtocol
import web

from ooresults.utils.globals import t_globals
from ooresults.handler import model
from ooresults.repo.repo import EventNotFoundError
from ooresults.repo import result_type
from ooresults.repo.result_type import ResultStatus

templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals=t_globals)


class WebSocketHandler:
    def __init__(self, demo_reader: bool = False):
        self.demo_reader = demo_reader
        self.connections = {}
        self.messages = []
        self.cardreader_status = {}
        self.executor = ThreadPoolExecutor(max_workers=2)

        #
        # Cardreader status:
        #
        # offline
        # online
        # readerDisconnected
        # readerConnected
        # cardInserted
        # cardRemoved
        # cardRead
        #

    async def update_event(self, event: Dict) -> None:
        if event:
            connections = {
                k: v
                for k, v in self.connections.items()
                if v[0] == "/si2" and str(event["id"]) == v[1]
            }
            if connections:
                await asyncio.wait(
                    [
                        self.send(conn=conn, event=copy.deepcopy(event), message={})
                        for conn in connections
                    ]
                )

    async def send_to_all(self, event: Dict, message: Dict = {}) -> None:
        if event:
            connections = {
                k: v for k, v in self.connections.items() if str(event["id"]) == v[1]
            }
            if connections:
                await asyncio.wait(
                    [
                        self.send(
                            conn=conn, event=copy.deepcopy(event), message=message
                        )
                        for conn in connections
                    ]
                )

    async def send(
        self, conn: WebSocketServerProtocol, event: Dict, message: Dict = {}
    ) -> None:
        status = (
            self.cardreader_status[event["id"]]
            if event and event["id"] in self.cardreader_status
            else "offline"
        )
        path, event_id, event_key = self.connections[conn]
        if path == "/si2":
            data = render.si.si2_data(status, event, self.messages)
        else:
            if status != "cardRead":
                data = message.get("controlCard", "")
            elif message.get("error", None) is not None:
                data = render.si.si1_error(message)
            elif message.get("lastName", None) is not None:
                data = render.si.si1_data(message)
            else:
                return
            data = json.dumps({"status": status, "data": str(data)})
        print("WEBSOCKET SEND ...", conn)
        await conn.send(str(data))

    async def handler(self, websocket: WebSocketServerProtocol, path: str) -> None:
        print("WEBSOCKET CONNECTED")
        print("websocket:", websocket)
        print("websocket.request_headers:", websocket.request_headers)
        print("path:", path)

        if path == "/demo":
            event = None
            try:
                if self.demo_reader:
                    async for message in websocket:
                        # workaround to detect lost websocket connection in the browser
                        # see https://stackoverflow.com/questions/26971026/handling-connection-loss-with-websockets
                        if message == "__ping__":
                            await websocket.send("__pong__")
                        else:
                            print("WEBSOCKET RECEIVED", websocket, message)

                            item = json.loads(message)
                            #
                            # item = {
                            #     'key': 4711,
                            #     'code': ['Check', 'Start', '', '', '', '', '', '', '', '', 'Finish'],
                            #     'time': ['10:11:12', '', '', '', '', '', '', '', '', '', '10:23:23'],
                            #     'card': '1111',
                            # }
                            #

                            d = {
                                "entryType": "cardRead",
                                "entryTime": datetime.datetime.now(),
                                "controlCard": item.get("card", None),
                                "result": None,
                            }

                            # add the event date to the times entered on the webpage
                            events = await asyncio.get_event_loop().run_in_executor(
                                executor=self.executor, func=model.get_events
                            )
                            date_of_event = datetime.date.today()
                            for e in events:
                                if e.key == item.get("key", None):
                                    date_of_event = e.date
                                    break

                            def parse_time(value: str) -> datetime.datetime:
                                return datetime.datetime.combine(
                                    date=date_of_event,
                                    time=datetime.datetime.strptime(
                                        value, "%H:%M:%S"
                                    ).time(),
                                    tzinfo=tzlocal.get_localzone(),
                                )

                            result = result_type.PersonRaceResult(
                                status=ResultStatus.FINISHED
                            )
                            _code = item["code"]
                            _time = item["time"]
                            if _code[0] == "Check" and _time[0] != "":
                                result.punched_check_time = parse_time(_time[0])
                            if _code[1] == "Start" and _time[1] != "":
                                result.punched_start_time = parse_time(_time[1])
                            if _code[-1] == "Finish" and _time[-1] != "":
                                result.punched_finish_time = parse_time(_time[-1])

                            result.start_time = result.punched_start_time
                            result.finish_time = result.punched_finish_time
                            for i in range(2, 10):
                                if _code[i] != "" and _time[i] != "":
                                    result.split_times.append(
                                        result_type.SplitTime(
                                            control_code=_code[i],
                                            punch_time=parse_time(_time[i]),
                                            status="Additional",
                                        )
                                    )
                            d["result"] = result

                            try:
                                (
                                    status,
                                    event,
                                    res,
                                ) = await asyncio.get_event_loop().run_in_executor(
                                    executor=self.executor,
                                    func=functools.partial(
                                        model.store_cardreader_result,
                                        event_key=item["key"],
                                        item=d,
                                    ),
                                )
                            except EventNotFoundError as e:
                                raise RuntimeError(str(e))
                            except Exception as e:
                                traceback.print_exc()
                                raise RuntimeError(str(e))

                            self.cardreader_status[event["id"]] = status

                            if "entryTime" in res:
                                res["entryTime"] = res["entryTime"].strftime("%H:%M:%S")

                            if status == "cardRead":
                                self.messages.append(res.copy())
                            await self.send_to_all(
                                event=event.copy(), message=res.copy()
                            )

                            res["readerStatus"] = status
                            res["event"] = event["name"]
                            if "status" in res:
                                res["status"] = res["status"].name
                            print(res)

            except websockets.exceptions.ConnectionClosed:
                pass
            finally:
                print("WEBSOCKET CONNECTION CLOSED", websocket)
                if event:
                    if event["id"] in self.cardreader_status:
                        del self.cardreader_status[event["id"]]
                    await self.send_to_all(event=event.copy(), message={})

        elif path == "/cardreader":
            event = None
            event_key = websocket.request_headers.get("X-Event-Key", "")
            try:
                print(">>>>>>>>>>> cardreader", event_key)
                async for message in websocket:
                    try:
                        data = bz2.decompress(message)
                    except:
                        raise RuntimeError("Data not bz2 encoded")

                    try:
                        item = json.loads(data.decode())
                    except:
                        raise RuntimeError("Data not json deserialisable")

                    try:
                        item = model.parse_cardreader_log(item=item)
                    except Exception as e:
                        raise RuntimeError(str(e))

                    try:
                        (
                            status,
                            event,
                            res,
                        ) = await asyncio.get_event_loop().run_in_executor(
                            executor=self.executor,
                            func=functools.partial(
                                model.store_cardreader_result,
                                event_key=event_key,
                                item=item,
                            ),
                        )
                    except EventNotFoundError as e:
                        raise RuntimeError(str(e))
                    except Exception as e:
                        traceback.print_exc()
                        raise RuntimeError(str(e))

                    self.cardreader_status[event["id"]] = status

                    if "entryTime" in res:
                        res["entryTime"] = res["entryTime"].strftime("%H:%M:%S")

                    if status == "cardRead":
                        self.messages.append(res.copy())
                    await self.send_to_all(event=event.copy(), message=res.copy())

                    res["readerStatus"] = status
                    res["event"] = event["name"]
                    if "status" in res:
                        res["status"] = res["status"].name
                    print(res)
                    await websocket.send(json.dumps(res))

            except websockets.exceptions.ConnectionClosed:
                pass
            except Exception as e:
                await websocket.send(str(e))
                print(str(e))
            finally:
                print("CARDREADER CONNECTION CLOSED", websocket)
                if event:
                    if event["id"] in self.cardreader_status:
                        del self.cardreader_status[event["id"]]
                    await self.send_to_all(event=event.copy(), message={})

        else:
            try:
                if path in ("/si1", "/si2"):
                    data = await websocket.recv()
                    event_id, event_key = data.split(",")

                    # check event key
                    events = await asyncio.get_event_loop().run_in_executor(
                        executor=self.executor, func=model.get_events
                    )
                    for e in events:
                        if str(e.id) == event_id and e.key == event_key:
                            self.connections[websocket] = (path, event_id, event_key)
                            await self.send(conn=websocket, event=e, message={})
                            async for message in websocket:
                                # workaround to detect lost websocket connection in the browser
                                # see https://stackoverflow.com/questions/26971026/handling-connection-loss-with-websockets
                                if message == "__ping__":
                                    await websocket.send("__pong__")
                                else:
                                    print("WEBSOCKET RECEIVED", websocket, message)
                                    break
                    else:
                        await websocket.send("__no_access__")
            except websockets.exceptions.ConnectionClosed:
                pass
            finally:
                if websocket in self.connections:
                    del self.connections[websocket]
                print("WEBSOCKET CONNECTION CLOSED", websocket)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional

from ooresults.repo.result_type import PersonRaceResult
from ooresults.repo.result_type import ResultStatus
from ooresults.pdf.pdf import PDF


def create_pdf(event: Dict, results: Dict, landscape: bool = False) -> bytes:
    # Instantiation of inherited class
    pdf = PDF(name=event.name, landscape=landscape)
    pdf.set_margin(margin=10)
    pdf.set_auto_page_break(auto=True, margin=15)
    pdf.add_page()
    nr_splittimes = 20 if landscape else 12

    def cell(w: int, h: Optional[int] = None, txt: str = "", align: str = "L") -> None:
        while pdf.get_string_width(txt) > w:
            txt = txt[:-1]
        pdf.cell(w=w, h=h, txt=txt, align=align)

    def format_time(time: int, status: ResultStatus) -> str:
        if status == ResultStatus.OK:
            return "{:d}:{:02d}".format(time // 60, time % 60)
        else:
            return pdf.MAP_STATUS[status]

    def format_splittime(time: Optional[int]) -> str:
        if time is not None:
            return "{:d}:{:02d}".format(time // 60, time % 60)
        else:
            return "-----"

    def pre(t1: str = "", t2: str = "", t3: str = "") -> None:
        pdf.set_font(style="B")
        cell(w=7, h=None, txt=t1, align="R")
        cell(w=2, h=None, txt="")
        cell(w=33, h=None, txt=t2, align="L")
        cell(w=12, h=None, txt=t3, align="R")
        pdf.set_font(style="")

    def t(a: Optional[datetime], b: Optional[datetime]) -> Optional[int]:
        if a is not None and b is not None:
            diff = b.replace(microsecond=0) - a.replace(microsecond=0)
            return int(diff.total_seconds())
        else:
            return None

    for i, class_results in enumerate(results):
        class_, ranked_results = class_results
        standard = class_.params.otype == "standard"

        if i > 0:
            # insert a page break if there is not enough space left on the
            # page for the class header, the table header and two table rows
            if pdf.will_page_break(height=22):
                pdf.add_page()
            else:
                pdf.ln()
                pdf.ln()

        pdf.set_font(family="Carlito", style="B", size=8)
        pdf.cell(txt=class_.name)

        # print possible voided legs
        if ranked_results and ranked_results[0]["result"] is not None:
            voided_legs = ranked_results[0]["result"].voided_legs()
            if voided_legs:
                pdf.cell(txt=f'(Voided legs: {", ".join(voided_legs)})')

        pdf.ln()
        # print list of control codes as header
        codes = []
        for i, j in enumerate(
            [
                t
                for t in ranked_results[0]["result"].split_times
                if t.status != "Additional"
            ]
        ):
            if standard:
                codes.append(f"{str(i+1)}({j.control_code})")
            else:
                codes.append(f"{str(i+1)}")
        codes.append("F")
        for i, j in enumerate(codes):
            if i % nr_splittimes == 0:
                if i >= nr_splittimes:
                    pdf.ln()
                pre()
            cell(w=10, h=None, txt=j, align="R")

        pdf.ln()
        pdf.ln()

        ranked = False
        for result in ranked_results:

            def get(d: Dict, key: str) -> str:
                s = d.get(key, "")
                return s if s is not None else ""

            class Result:
                def __init__(self, result: PersonRaceResult):
                    self.result = result

                def next(self) -> str:
                    if (
                        self.result.start_time is not None
                        and self.result.finish_time is not None
                    ):
                        running_time = t(
                            self.result.start_time, self.result.finish_time
                        )
                    else:
                        running_time = self.result.extensions.get(
                            "running_time", self.result.time
                        )

                    last_time = 0
                    for i in [
                        t for t in self.result.split_times if t.status != "Additional"
                    ]:
                        if last_time is not None and i.time is not None:
                            split_time = format_splittime(time=int(i.time - last_time))
                            last_time = i.time
                        else:
                            split_time = "-----"
                        if standard:
                            if i.leg_voided:
                                split_time = "[" + split_time + "]"
                            yield format_splittime(i.time), split_time, None
                        else:
                            yield f"#({i.control_code})", format_splittime(
                                i.time
                            ), split_time

                    if last_time is not None and running_time is not None:
                        split_time = format_splittime(
                            time=int(running_time - last_time)
                        )
                    else:
                        split_time = "-----"
                    if standard:
                        if self.result.last_leg_voided:
                            split_time = "[" + split_time + "]"
                        yield format_splittime(running_time), split_time, None
                    else:
                        yield "F", format_splittime(running_time), split_time
                    additional = [
                        t for t in self.result.split_times if t.status == "Additional"
                    ]
                    if additional != []:
                        yield "", "", ""
                    for i in additional:
                        if standard:
                            yield format_splittime(i.time), f"*({i.control_code})", None
                        else:
                            yield f"*({i.control_code})", format_splittime(i.time), ""

            if result["result"].status not in (
                ResultStatus.INACTIVE,
                ResultStatus.DID_NOT_START,
            ):
                pdf.set_font(family="Carlito", size=8)
                if result.rank is not None:
                    ranked = True
                elif ranked:
                    ranked = False
                    pdf.ln()

                sp = 0
                result_data = Result(result=result["result"])

                def print_line_1(sp: int, line_1: List[str]) -> None:
                    if sp <= nr_splittimes:
                        running_time = result["result"].extensions.get(
                            "running_time", result["result"].time
                        )

                        pre(
                            t1=str(get(result, "rank"))
                            if not get(result, "not_competing")
                            else "AK",
                            t2=get(result, "last_name")
                            + " "
                            + get(result, "first_name"),
                            t3=format_time(running_time, result["result"].status),
                        )
                    else:
                        pdf.ln()
                        pre()

                    for i in line_1:
                        cell(w=10, h=None, txt=i, align="R")

                def print_line_2(sp: int, line_2: List[str]) -> None:
                    if sp <= nr_splittimes:
                        pdf.ln()
                        # first line: rank, lastname+firstname, time, splittimes
                        pre(t2=get(result, "club"))
                    else:
                        pdf.ln()
                        pre()

                    for i in line_2:
                        cell(w=10, h=None, txt=i, align="R")

                def print_line_3(sp: int, line_3: List[str]) -> None:
                    pdf.ln()
                    pre()

                    for i in line_3:
                        cell(w=10, h=None, txt=i, align="R")

                line_1 = []
                line_2 = []
                line_3 = []
                for i, j, k in result_data.next():
                    line_1.append(i)
                    line_2.append(j)
                    line_3.append(k)
                    sp += 1
                    if sp % nr_splittimes == 0:
                        print_line_1(sp=sp, line_1=line_1)
                        print_line_2(sp=sp, line_2=line_2)
                        if not standard:
                            print_line_3(sp=sp, line_3=line_3)
                        line_1 = []
                        line_2 = []
                        line_3 = []
                if line_1 != []:
                    print_line_1(sp=sp, line_1=line_1)
                    print_line_2(sp=sp, line_2=line_2)
                    if not standard:
                        print_line_3(sp=sp, line_3=line_3)
                pdf.ln()

    return bytes(pdf.output())



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import copy
import logging
import pathlib
import io
from decimal import Decimal

import web
import clevercsv as csv

from ooresults.handler import model
from ooresults.handler import results
from ooresults.repo import series_type
import ooresults.pdf.series


templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals={"str": str})


def build_total_results(settings: series_type.Settings, list_of_results, organizers=[]):
    q = Decimal(10) ** -settings.decimal_places

    r = {}
    for i, class_results in enumerate(list_of_results):
        for class_, results in class_results:
            if class_.name == "Organizer":
                continue
            if class_.name not in r:
                r[class_.name] = {}
            for entry in results:
                # check if competitor has points
                if entry.get("points", None) is not None:
                    if (entry["last_name"], entry["first_name"]) not in r[class_.name]:
                        r[class_.name][(entry["last_name"], entry["first_name"])] = {
                            "last_name": entry["last_name"],
                            "first_name": entry["first_name"],
                            "year": entry.get("year", None),
                            "club": entry.get("club", None),
                            "sum": 0,
                            "races": {},
                            "organizer": {},
                        }

                    p = Decimal(settings.maximum_points * entry["points"]).quantize(q)
                    r[class_.name][(entry["last_name"], entry["first_name"])]["races"][
                        i
                    ] = p
            # add organizers
            if len(organizers) > i:
                for entry in organizers[i]:
                    if (entry["last_name"], entry["first_name"]) not in r[class_.name]:
                        r[class_.name][(entry["last_name"], entry["first_name"])] = {
                            "last_name": entry["last_name"],
                            "first_name": entry["first_name"],
                            "year": entry.get("year", None),
                            "club": entry.get("club", None),
                            "sum": 0,
                            "races": {},
                            "organizer": {},
                        }
                    r[class_.name][(entry["last_name"], entry["first_name"])][
                        "organizer"
                    ][i] = 0

    # build sum of points
    ranked_classes = []
    for class_name, entries in r.items():
        for entry in entries.values():
            points_of_series = sorted(entry["races"].values(), reverse=True)
            # add organizer bonus
            if entry["organizer"] != {}:
                if len(points_of_series) == 0:
                    points = Decimal(0)
                elif len(points_of_series) == 1:
                    points = points_of_series[0] / 2
                else:
                    points = (points_of_series[0] + points_of_series[1]) / 2
                for i in entry["organizer"].keys():
                    entry["organizer"][i] = points.quantize(q)

            # compute sum
            points_of_series = sorted(
                list(entry["races"].values()) + list(entry["organizer"].values()),
                reverse=True,
            )
            if settings.nr_of_best_results is not None:
                points_of_series = points_of_series[0 : settings.nr_of_best_results]
            for points in points_of_series:
                entry["sum"] += points

        # build a list and rank the list
        ranked_entries = list(entries.values())
        ranked_entries.sort(key=lambda e: e["last_name"] + "," + e["first_name"])
        ranked_entries.sort(key=lambda e: e["sum"], reverse=True)

        for j, e in enumerate(ranked_entries):
            if j > 0 and ranked_entries[j]["sum"] == ranked_entries[j - 1]["sum"]:
                e["rank"] = ranked_entries[j - 1]["rank"]
            else:
                e["rank"] = j + 1
            # rank is None if sum is 0
            if e["sum"] == 0:
                e["rank"] = None

        ranked_classes.append((class_name, ranked_entries))

    return ranked_classes


def create_event_list(events):
    # filter list
    e_list = [e for e in events if e.series is not None]
    # sort list
    e_list.sort(key=lambda e: e.series)
    e_list.sort(key=lambda e: e.date)
    return e_list


def build_result():
    settings = model.get_series_settings()
    # build event list
    events = list(model.get_events())
    events = create_event_list(events=events)

    list_of_results = []
    organizers = []
    for i, event in enumerate(events):
        classes = list(model.get_classes(event_id=event.id))
        entry_list = list(model.get_entries(event_id=event.id))
        class_results = results.build_results(
            classes=classes, results=copy.deepcopy(entry_list)
        )
        list_of_results.append(class_results)
        organizers.append([e for e in entry_list if e.class_ == "Organizer"])

    return (
        settings,
        events,
        build_total_results(
            settings=settings, list_of_results=list_of_results, organizers=organizers
        ),
    )


class Update:
    def POST(self):
        """Update data"""
        data = web.input()
        settings, events, results = build_result()
        return render.series_table(events, results)


class Settings:
    def POST(self):
        """Update seeries settings"""
        data = web.input()
        print(data)
        try:
            settings = series_type.Settings(
                name=data.name,
                nr_of_best_results=int(data.nr_of_best_results)
                if data.nr_of_best_results != ""
                else None,
                mode=data.mode,
                maximum_points=int(data.maximum_points),
                decimal_places=int(data.decimal_places),
            )
            model.update_series_settings(settings=settings)
        except Exception as e:
            raise web.internalerror(str(e))

        settings, events, results = build_result()
        return render.series_table(events, results)


class PdfResult:
    def POST(self):
        """Print results"""
        data = web.input()
        landscape = "ser_landscape" in data

        try:
            settings, events, results = build_result()
            content = ooresults.pdf.series.create_pdf(
                settings=settings, events=events, results=results, landscape=landscape
            )
            return content

        except KeyError:
            raise web.conflict("Internal error")
        except:
            logging.exception("Internal server error")
            raise


class CsvResult:
    def POST(self):
        """Export results as csv for creating diplomas"""
        data = web.input()
        try:
            settings, events, results = build_result()

            output = io.StringIO()
            writer = csv.writer(output, delimiter=";", quoting=csv.QUOTE_MINIMAL)

            # write header
            writer.writerow(
                [
                    "Kategorie",
                    "Pos",
                    "Pos_eng",
                    "Vorname",
                    "Nachname",
                    "Verein",
                    "Punkte",
                ]
            )

            for class_name, ranked_results in results:
                for result in ranked_results:
                    if result.get("rank", None) is not None:
                        rank = str(result.get("rank", None))
                        if rank == "1":
                            rank_eng = rank + "st"
                        elif rank == "2":
                            rank_eng = rank + "nd"
                        elif rank == "3":
                            rank_eng = rank + "rd"
                        else:
                            rank_eng = rank + "th"
                        writer.writerow(
                            [
                                class_name,
                                rank,
                                rank_eng,
                                result.get("first_name", ""),
                                result.get("last_name", ""),
                                result.get("club", ""),
                                str(result.get("sum", None)),
                            ]
                        )

            content = output.getvalue()
            output.close()
            return content.encode(encoding="utf-8")

        except KeyError:
            raise web.conflict("Internal error")
        except:
            logging.exception("Internal server error")
            raise


class FillSettingsForm:
    def POST(self):
        """Query data to fill settings form"""
        settings = model.get_series_settings()
        return render.series_settings(settings)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pathlib
from typing import List
from typing import Dict
from typing import Tuple

from lxml import etree
from lxml.builder import ElementMaker


schema_file = pathlib.Path(__file__).parent.parent / "schema" / "IOF.xsd"
xml_schema = etree.XMLSchema(etree.parse(str(schema_file)))


iof_namespace = "http://www.orienteering.org/datastandard/3.0"
namespaces = {None: iof_namespace}


def create_class_list(classes: List[Dict]) -> bytes:
    E = ElementMaker(namespace=iof_namespace, nsmap=namespaces)

    CLASSLIST = E.ClassList
    CLASS = E.Class
    NAME = E.Name
    SHORTNAME = E.ShortName

    root = CLASSLIST(iofVersion="3.0")

    for c in classes:
        class_ = CLASS()
        class_.append(NAME(c["name"]))

        if c.get("short_name", None) is not None:
            class_.append(SHORTNAME(str(c["short_name"])))

        root.append(class_)

    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    return etree.tostring(
        root, encoding="UTF-8", xml_declaration=True, pretty_print=True
    )


def parse_class_list(content: bytes) -> Tuple[List[Dict]]:
    # Class: Dict of {'name': str, 'short_name': Optional[str]}

    root = etree.XML(content)
    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    if not root.tag == "{" + iof_namespace + "}ClassList":
        raise RuntimeError("Root element is " + root.tag + " but should be ClassList")

    classes = []
    class_list = root.findall("Class", namespaces=namespaces)

    for class_ in class_list:
        c = {
            "name": "",
            "short_name": None,
        }
        c["name"] = class_.find("Name", namespaces=namespaces).text

        short_name = class_.find("ShortName", namespaces=namespaces)
        if short_name is not None:
            c["short_name"] = short_name.text

        classes.append(c)

    return classes



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from datetime import datetime
from datetime import timedelta
from datetime import timezone

import pytest
import jsonschema

from ooresults.repo.result_type import SplitTime
from ooresults.repo.result_type import ResultStatus
from ooresults.handler import model


def test_parse_with_start_time_and_with_finish_time():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "startTime": "2015-01-01T12:38:59Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
        "finishTime": "2015-01-01T12:39:07Z",
    }

    d = model.parse_cardreader_log(item=item)
    assert d["entryType"] == "cardRead"
    assert d["entryTime"] == e1
    assert d["controlCard"] == "123"
    assert d["result"].start_time == s1
    assert d["result"].finish_time == f1
    assert d["result"].punched_clear_time == None
    assert d["result"].punched_check_time == None
    assert d["result"].punched_start_time == s1
    assert d["result"].punched_finish_time == f1
    assert d["result"].time == None
    assert d["result"].status == ResultStatus.FINISHED
    assert len(d["result"].split_times) == 2
    assert d["result"].split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=None, status="Additional"
    )
    assert d["result"].split_times[1] == SplitTime(
        control_code="103", punch_time=c2, time=None, status="Additional"
    )


def test_parse_without_start_time_and_with_finish_time():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
        "finishTime": "2015-01-01T12:39:07Z",
    }

    d = model.parse_cardreader_log(item=item)
    assert d["entryType"] == "cardRead"
    assert d["entryTime"] == e1
    assert d["controlCard"] == "123"
    assert d["result"].start_time == None
    assert d["result"].finish_time == f1
    assert d["result"].punched_clear_time == None
    assert d["result"].punched_check_time == None
    assert d["result"].punched_start_time == None
    assert d["result"].punched_finish_time == f1
    assert d["result"].time == None
    assert d["result"].status == ResultStatus.FINISHED
    assert len(d["result"].split_times) == 2
    assert d["result"].split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=None, status="Additional"
    )
    assert d["result"].split_times[1] == SplitTime(
        control_code="103", punch_time=c2, time=None, status="Additional"
    )


def test_parse_with_start_time_and_without_finish_time():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "startTime": "2015-01-01T12:38:59Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
    }

    d = model.parse_cardreader_log(item=item)
    assert d["entryType"] == "cardRead"
    assert d["entryTime"] == e1
    assert d["controlCard"] == "123"
    assert d["result"].start_time == s1
    assert d["result"].finish_time == None
    assert d["result"].punched_clear_time == None
    assert d["result"].punched_check_time == None
    assert d["result"].punched_start_time == s1
    assert d["result"].punched_finish_time == None
    assert d["result"].time == None
    assert d["result"].status == ResultStatus.FINISHED
    assert len(d["result"].split_times) == 2
    assert d["result"].split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=None, status="Additional"
    )
    assert d["result"].split_times[1] == SplitTime(
        control_code="103", punch_time=c2, time=None, status="Additional"
    )


def test_parse_without_start_time_and_without_finish_time():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
    }

    d = model.parse_cardreader_log(item=item)
    assert d["entryType"] == "cardRead"
    assert d["entryTime"] == e1
    assert d["controlCard"] == "123"
    assert d["result"].start_time == None
    assert d["result"].finish_time == None
    assert d["result"].punched_clear_time == None
    assert d["result"].punched_check_time == None
    assert d["result"].punched_start_time == None
    assert d["result"].punched_finish_time == None
    assert d["result"].time == None
    assert d["result"].status == ResultStatus.FINISHED
    assert len(d["result"].split_times) == 2
    assert d["result"].split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=None, status="Additional"
    )
    assert d["result"].split_times[1] == SplitTime(
        control_code="103", punch_time=c2, time=None, status="Additional"
    )


def test_parse_with_clear_time():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    cl = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "clearTime": "2015-01-01T12:38:59Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
    }

    d = model.parse_cardreader_log(item=item)
    assert d["entryType"] == "cardRead"
    assert d["entryTime"] == e1
    assert d["controlCard"] == "123"
    assert d["result"].start_time == None
    assert d["result"].finish_time == None
    assert d["result"].punched_clear_time == cl
    assert d["result"].punched_check_time == None
    assert d["result"].punched_start_time == None
    assert d["result"].punched_finish_time == None
    assert d["result"].time == None
    assert d["result"].status == ResultStatus.FINISHED
    assert len(d["result"].split_times) == 2
    assert d["result"].split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=None, status="Additional"
    )
    assert d["result"].split_times[1] == SplitTime(
        control_code="103", punch_time=c2, time=None, status="Additional"
    )


def test_parse_with_check_time():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    ch = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "checkTime": "2015-01-01T12:38:59Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
    }

    d = model.parse_cardreader_log(item=item)
    assert d["entryType"] == "cardRead"
    assert d["entryTime"] == e1
    assert d["controlCard"] == "123"
    assert d["result"].start_time == None
    assert d["result"].finish_time == None
    assert d["result"].punched_clear_time == None
    assert d["result"].punched_check_time == ch
    assert d["result"].punched_start_time == None
    assert d["result"].punched_finish_time == None
    assert d["result"].time == None
    assert d["result"].status == ResultStatus.FINISHED
    assert len(d["result"].split_times) == 2
    assert d["result"].split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=None, status="Additional"
    )
    assert d["result"].split_times[1] == SplitTime(
        control_code="103", punch_time=c2, time=None, status="Additional"
    )


def test_parse_without_controls():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "startTime": "2015-01-01T12:38:59Z",
        "controlCard": "123",
        "punches": [],
        "finishTime": "2015-01-01T12:39:07Z",
    }

    d = model.parse_cardreader_log(item=item)
    assert d["entryType"] == "cardRead"
    assert d["entryTime"] == e1
    assert d["controlCard"] == "123"
    assert d["result"].start_time == s1
    assert d["result"].finish_time == f1
    assert d["result"].punched_clear_time == None
    assert d["result"].punched_check_time == None
    assert d["result"].punched_start_time == s1
    assert d["result"].punched_finish_time == f1
    assert d["result"].time == None
    assert d["result"].status == ResultStatus.FINISHED
    assert len(d["result"].split_times) == 0
    assert d["result"].split_times == []


def test_parse_with_many_controls():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 3, tzinfo=timezone.utc)
    c3 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    c4 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)
    c5 = datetime(2015, 1, 1, 12, 39, 9, tzinfo=timezone.utc)
    c6 = datetime(2015, 1, 1, 12, 39, 11, tzinfo=timezone.utc)
    c7 = datetime(2015, 1, 1, 12, 39, 13, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 15, tzinfo=timezone.utc)

    controls = ["101", "102", "103"]
    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "startTime": "2015-01-01T12:38:59Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "105", "punchTime": "2015-01-01T12:39:03Z"},
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:05Z"},
            {"controlCode": "102", "punchTime": "2015-01-01T12:39:07Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:09Z"},
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:11Z"},
            {"controlCode": "104", "punchTime": "2015-01-01T12:39:13Z"},
        ],
        "finishTime": "2015-01-01T12:39:15Z",
    }

    d = model.parse_cardreader_log(item=item)
    assert d["entryType"] == "cardRead"
    assert d["entryTime"] == e1
    assert d["controlCard"] == "123"
    assert d["result"].start_time == s1
    assert d["result"].finish_time == f1
    assert d["result"].punched_clear_time == None
    assert d["result"].punched_check_time == None
    assert d["result"].punched_start_time == s1
    assert d["result"].punched_finish_time == f1
    assert d["result"].time == None
    assert d["result"].status == ResultStatus.FINISHED
    assert len(d["result"].split_times) == 7
    assert d["result"].split_times[0] == SplitTime(
        control_code="101", punch_time=c1, time=None, status="Additional"
    )
    assert d["result"].split_times[1] == SplitTime(
        control_code="105", punch_time=c2, time=None, status="Additional"
    )
    assert d["result"].split_times[2] == SplitTime(
        control_code="101", punch_time=c3, time=None, status="Additional"
    )
    assert d["result"].split_times[3] == SplitTime(
        control_code="102", punch_time=c4, time=None, status="Additional"
    )
    assert d["result"].split_times[4] == SplitTime(
        control_code="103", punch_time=c5, time=None, status="Additional"
    )
    assert d["result"].split_times[5] == SplitTime(
        control_code="101", punch_time=c6, time=None, status="Additional"
    )
    assert d["result"].split_times[6] == SplitTime(
        control_code="104", punch_time=c7, time=None, status="Additional"
    )


def test_parse_without_entry_type_raises_exception():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)

    item = {
        "entryTime": "2015-01-01T13:00:00Z",
        "startTime": "2015-01-01T12:38:59Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
        "finishTime": "2015-01-01T12:39:07Z",
    }

    with pytest.raises(jsonschema.ValidationError):
        d = model.parse_cardreader_log(item=item)


def test_parse_without_entry_time_raises_exception():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "startTime": "2015-01-01T12:38:59Z",
        "controlCard": "123",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
        "finishTime": "2015-01-01T12:39:07Z",
    }

    with pytest.raises(jsonschema.ValidationError):
        d = model.parse_cardreader_log(item=item)


def test_parse_without_control_card_raises_exception():
    e1 = datetime(2015, 1, 1, 13, 0, 0, tzinfo=timezone.utc)
    s1 = datetime(2015, 1, 1, 12, 38, 59, tzinfo=timezone.utc)
    c1 = datetime(2015, 1, 1, 12, 39, 1, tzinfo=timezone.utc)
    c2 = datetime(2015, 1, 1, 12, 39, 5, tzinfo=timezone.utc)
    f1 = datetime(2015, 1, 1, 12, 39, 7, tzinfo=timezone.utc)

    item = {
        "entryType": "cardRead",
        "entryTime": "2015-01-01T13:00:00Z",
        "startTime": "2015-01-01T12:38:59Z",
        "punches": [
            {"controlCode": "101", "punchTime": "2015-01-01T12:39:01Z"},
            {"controlCode": "103", "punchTime": "2015-01-01T12:39:05Z"},
        ],
        "finishTime": "2015-01-01T12:39:07Z",
    }

    with pytest.raises(jsonschema.ValidationError):
        d = model.parse_cardreader_log(item=item)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

import pytest

from ooresults.repo import repo
from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.class_params import ClassParams


@pytest.fixture
def db():
    return SqliteRepo(db=":memory:")


@pytest.fixture
def event_id(db):
    return db.add_event(
        name="Event",
        date=datetime.date(year=2020, month=1, day=1),
        key=None,
        publish=False,
        series=True,
    )


@pytest.fixture
def course_1_id(db, event_id):
    return db.add_course(
        event_id=event_id,
        name="Course 1",
        length=4500,
        climb=90,
        controls=["101", "102", "103"],
    )


@pytest.fixture
def course_2_id(db, event_id):
    return db.add_course(
        event_id=event_id,
        name="Course 2",
        length=None,
        climb=None,
        controls=[],
    )


@pytest.fixture
def class_id(db, event_id, course_1_id):
    return db.add_class(
        event_id=event_id,
        name="Class 1",
        short_name=None,
        course_id=course_1_id,
        params=ClassParams(),
    )


def test_get_courses_after_adding_one_course(db, event_id, course_1_id):
    c = list(db.get_courses(event_id=event_id))
    assert len(c) == 1
    assert c[0].id == course_1_id
    assert c[0].name == "Course 1"
    assert c[0].length == 4500
    assert c[0].climb == 90
    assert c[0].controls == ["101", "102", "103"]


def test_get_courses_after_adding_two_courses(db, event_id, course_1_id, course_2_id):
    c = list(db.get_courses(event_id=event_id))
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == course_1_id
    assert c[0].name == "Course 1"
    assert c[0].length == 4500
    assert c[0].climb == 90
    assert c[0].controls == ["101", "102", "103"]
    assert c[1].id == course_2_id
    assert c[1].name == "Course 2"
    assert c[1].length == None
    assert c[1].climb == None
    assert c[1].controls == []


def test_get_first_added_course(db, course_1_id, course_2_id):
    c = list(db.get_course(id=course_1_id))
    assert len(c) == 1
    assert c[0].id == course_1_id
    assert c[0].name == "Course 1"
    assert c[0].length == 4500
    assert c[0].climb == 90
    assert c[0].controls == ["101", "102", "103"]


def test_get_last_added_course(db, course_1_id, course_2_id):
    c = list(db.get_course(id=course_2_id))
    assert len(c) == 1
    assert c[0].id == course_2_id
    assert c[0].name == "Course 2"
    assert c[0].length == None
    assert c[0].climb == None
    assert c[0].controls == []


def test_update_first_added_course(db, event_id, course_1_id, course_2_id):
    db.update_course(
        id=course_1_id, name="Course 3", length=3900, climb=150, controls=["101"]
    )
    c = list(db.get_courses(event_id=event_id))
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == course_2_id
    assert c[0].name == "Course 2"
    assert c[0].length == None
    assert c[0].climb == None
    assert c[0].controls == []
    assert c[1].id == course_1_id
    assert c[1].name == "Course 3"
    assert c[1].length == 3900
    assert c[1].climb == 150
    assert c[1].controls == ["101"]


def test_update_last_added_course(db, event_id, course_1_id, course_2_id):
    db.update_course(
        id=course_2_id,
        name="Course 3",
        length=5100,
        climb=None,
        controls=["301", "302", "303"],
    )
    c = list(db.get_courses(event_id=event_id))
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == course_1_id
    assert c[0].name == "Course 1"
    assert c[0].length == 4500
    assert c[0].climb == 90
    assert c[0].controls == ["101", "102", "103"]
    assert c[1].id == course_2_id
    assert c[1].name == "Course 3"
    assert c[1].length == 5100
    assert c[1].climb == None
    assert c[1].controls == ["301", "302", "303"]


def test_delete_first_added_course(db, event_id, course_1_id, course_2_id):
    db.delete_course(id=course_1_id)
    c = list(db.get_courses(event_id=event_id))
    assert len(c) == 1
    assert c[0].id == course_2_id
    assert c[0].name == "Course 2"
    assert c[0].length == None
    assert c[0].climb == None
    assert c[0].controls == []


def test_delete_last_added_course(db, event_id, course_1_id, course_2_id):
    db.delete_course(id=course_2_id)
    c = list(db.get_courses(event_id=event_id))
    assert len(c) == 1
    assert c[0].id == course_1_id
    assert c[0].name == "Course 1"
    assert c[0].length == 4500
    assert c[0].climb == 90
    assert c[0].controls == ["101", "102", "103"]


def test_delete_courses_deletes_all_courses(db, event_id, course_1_id, course_2_id):
    db.delete_courses(event_id=event_id)
    c = list(db.get_courses(event_id=event_id))
    assert len(c) == 0


def test_add_existing_name_raises_exception(db, event_id, course_1_id):
    with pytest.raises(repo.ConstraintError, match="Course already exist"):
        db.add_course(
            event_id=event_id, name="Course 1", length=None, climb=None, controls=[]
        )


def test_change_to_existing_name_raises_exception(db, course_1_id, course_2_id):
    with pytest.raises(repo.ConstraintError, match="Course already exist"):
        db.update_course(
            id=course_1_id, name="Course 2", length=None, climb=None, controls=[]
        )


def test_update_with_unknown_id_raises_exception(db, course_1_id):
    with pytest.raises(KeyError):
        db.update_course(
            id=course_1_id + 1, name="Course 2", length=None, climb=None, controls=[]
        )


def test_add_course_with_unknown_event_id_raises_exception(db, event_id):
    with pytest.raises(repo.EventNotFoundError):
        db.add_course(
            event_id=event_id + 1, name="Course 1", length=None, climb=None, controls=[]
        )


def test_delete_course_with_unknown_id_do_not_change_anything(
    db, event_id, course_1_id
):
    db.delete_course(id=course_1_id + 1)
    c = list(db.get_courses(event_id=event_id))
    assert len(c) == 1
    assert c[0].id == course_1_id
    assert c[0].name == "Course 1"
    assert c[0].length == 4500
    assert c[0].climb == 90
    assert c[0].controls == ["101", "102", "103"]


def test_delete_course_used_in_class_raises_exception(
    db, event_id, class_id, course_1_id
):
    with pytest.raises(repo.CourseUsedError):
        db.delete_course(id=course_1_id)


def test_delete_courses_used_in_class_raises_exception(
    db, event_id, class_id, course_1_id
):
    with pytest.raises(repo.CourseUsedError):
        db.delete_courses(event_id=event_id)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from datetime import datetime
from datetime import timezone

from ooresults.plugins.oe2003 import parse
from ooresults.repo import result_type
from ooresults.repo import start_type
from ooresults.repo.result_type import ResultStatus


header = [
    "Chip",
    "Vorname",
    "Nachname",
    "Start",
    "Ziel",
    "Zeit",
    "G",
    "Jg",
    "AK",
    "Wertung",
    "Lang",
    "Ort",
]
tz = datetime.now(timezone.utc).astimezone().tzinfo


def test_separator_comma():
    value = "c,v,n,,,,,,,,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_separator_semicolon():
    value = "c;v;n;;;;;;;;;"
    content = bytes(";".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_separator_tab():
    value = "c\tv\tn\t\t\t\t\t\t\t\t\t"
    content = bytes("\t".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_quoted_data():
    value = '"c","v","n","","","","","","","","",""'
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_quote_within_quotes():
    value = 'c,v,"n1""2",,,,,,,,,'
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": 'n1"2',
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_separator_within_quotes():
    value = 'c,v,"n1,2",,,,,,,,,'
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n1,2",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_newline_within_quotes():
    value = 'c,v,"n1\n2",,,,,,,,,'
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n1\n2",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_status_ok_and_defined_time():
    value = "c,v,n,,,0:10,,,,0,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(status=ResultStatus.OK, time=10),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_status_ok_but_no_time():
    value = "c,v,n,,,,,,,0,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_status_dns():
    value = "c,v,n,,,,,,,1,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DID_NOT_START, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_status_dnf():
    value = "c,v,n,,,,,,,2,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DID_NOT_FINISH, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_status_mp():
    value = "c,v,n,,,,,,,3,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.MISSING_PUNCH, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_status_disq():
    value = "c,v,n,,,,,,,4,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DISQUALIFIED, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_status_over_time():
    value = "c,v,n,,,,,,,5,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.OVER_TIME, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_gender_male():
    value = "c,v,n,,,,M,,,4,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "M",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DISQUALIFIED, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_gender_female():
    value = "c,v,n,,,,F,,,4,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "F",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DISQUALIFIED, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_year2():
    value = "c,v,n,,,,,15,,4,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": 2015,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DISQUALIFIED, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_year4():
    value = "c,v,n,,,,,1915,,4,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": 1915,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DISQUALIFIED, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_not_competing_is_false():
    value = "c,v,n,,,,,,0,4,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DISQUALIFIED, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_not_competing_is_true():
    value = "c,v,n,,,,,,X,4,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": True,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DISQUALIFIED, time=None
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_start_time():
    value = "c,v,n,14:05:00,,,,,,,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(
                start_time=datetime(1900, 1, 1, 14, 5, 0)
            ),
        }
    ]


def test_relative_start_time_h_m_s():
    value = "c,v,n,4:05:00,,,,,,,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(
                start_time=datetime(1900, 1, 1, 4, 5, 0)
            ),
        }
    ]


def test_relative_start_time_m_s():
    value = "c,v,n,245:02,,,,,,,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE, time=None
            ),
            "start": start_type.PersonRaceStart(
                start_time=datetime(1900, 1, 1, 4, 5, 2)
            ),
        }
    ]


def test_finish_time():
    value = "c,v,n,,18:59:12,,,,,,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE,
                finish_time=datetime(1900, 1, 1, 18, 59, 12),
                punched_finish_time=datetime(1900, 1, 1, 18, 59, 12),
                time=None,
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_relative_finish_time_h_m_s():
    value = "c,v,n,,1:59:12,,,,,,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE,
                finish_time=datetime(1900, 1, 1, 1, 59, 12),
                punched_finish_time=datetime(1900, 1, 1, 1, 59, 12),
                time=None,
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_relative_finish_time_m_s():
    value = "c,v,n,,299:12,,,,,,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE,
                finish_time=datetime(1900, 1, 1, 4, 59, 12),
                punched_finish_time=datetime(1900, 1, 1, 4, 59, 12),
                time=None,
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_start_and_finish_time():
    value = "c,v,n,14:05:00,14:36:11,,,,,,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE,
                start_time=datetime(1900, 1, 1, 14, 5, 0),
                punched_start_time=datetime(1900, 1, 1, 14, 5, 0),
                finish_time=datetime(1900, 1, 1, 14, 36, 11),
                punched_finish_time=datetime(1900, 1, 1, 14, 36, 11),
                time=None,
            ),
            "start": start_type.PersonRaceStart(
                start_time=datetime(1900, 1, 1, 14, 5, 0)
            ),
        }
    ]


def test_start_time_and_status_mp():
    value = "c,v,n,14:05:00,,,,,,3,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.MISSING_PUNCH,
                start_time=datetime(1900, 1, 1, 14, 5, 0),
                punched_start_time=datetime(1900, 1, 1, 14, 5, 0),
                time=None,
            ),
            "start": start_type.PersonRaceStart(
                start_time=datetime(1900, 1, 1, 14, 5, 0)
            ),
        }
    ]


def test_start_time_and_status_dnf():
    value = "c,v,n,14:05:00,,,,,,2,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.DID_NOT_FINISH,
                start_time=datetime(1900, 1, 1, 14, 5, 0),
                punched_start_time=datetime(1900, 1, 1, 14, 5, 0),
                time=None,
            ),
            "start": start_type.PersonRaceStart(
                start_time=datetime(1900, 1, 1, 14, 5, 0)
            ),
        }
    ]


def test_start_time_and_status_over_time():
    value = "c,v,n,14:05:00,,,,,,5,,"
    content = bytes(",".join(header) + "\n" + value, encoding="utf-8")
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.OVER_TIME,
                start_time=datetime(1900, 1, 1, 14, 5, 0),
                punched_start_time=datetime(1900, 1, 1, 14, 5, 0),
                time=None,
            ),
            "start": start_type.PersonRaceStart(
                start_time=datetime(1900, 1, 1, 14, 5, 0)
            ),
        }
    ]


def test_split_time():
    value = "c,v,n,,,,,,,,,,123,4:16"
    content = bytes(
        ",".join(header + ["Posten1,Punch1"]) + "\n" + value, encoding="utf-8"
    )
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE,
                split_times=[
                    result_type.SplitTime(control_code="123", time=256, status=None),
                ],
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_split_time_without_punch():
    value = "c,v,n,,,,,,,,,,123,-----"
    content = bytes(
        ",".join(header + ["Posten1,Punch1"]) + "\n" + value, encoding="utf-8"
    )
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE,
                split_times=[
                    result_type.SplitTime(
                        control_code="123", time=None, status="Missing"
                    ),
                ],
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_two_split_times():
    value = "c,v,n,,,,,,,,,,123,4:16,99,12:01"
    content = bytes(
        ",".join(header + ["Posten1,Punch1"]) + "\n" + value, encoding="utf-8"
    )
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE,
                split_times=[
                    result_type.SplitTime(control_code="123", time=256, status=None),
                    result_type.SplitTime(control_code="99", time=721, status=None),
                ],
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_split_times_with_closing_separator():
    value = "c,v,n,,,,,,,,,,123,4:16,99,12:01,"
    content = bytes(
        ",".join(header + ["Posten1,Punch1"]) + "\n" + value, encoding="utf-8"
    )
    assert parse(content) == [
        {
            "first_name": "v",
            "last_name": "n",
            "class_": "",
            "club": "",
            "chip": "c",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(
                status=ResultStatus.INACTIVE,
                split_times=[
                    result_type.SplitTime(control_code="123", time=256, status=None),
                    result_type.SplitTime(control_code="99", time=721, status=None),
                ],
            ),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_import_several_lines():
    value_1 = "c1,v1,n1,,,0:11,,,,0,,"
    value_2 = "c2,v2,n2,,,0:12,,,,0,,"
    value_3 = "c3,v3,n3,,,0:13,,,,0,,"
    content = bytes(
        ",".join(header) + "\n" + value_1 + "\n" + value_2 + "\n" + value_3,
        encoding="utf-8",
    )
    assert parse(content) == [
        {
            "first_name": "v1",
            "last_name": "n1",
            "class_": "",
            "club": "",
            "chip": "c1",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(status=ResultStatus.OK, time=11),
            "start": start_type.PersonRaceStart(),
        },
        {
            "first_name": "v2",
            "last_name": "n2",
            "class_": "",
            "club": "",
            "chip": "c2",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(status=ResultStatus.OK, time=12),
            "start": start_type.PersonRaceStart(),
        },
        {
            "first_name": "v3",
            "last_name": "n3",
            "class_": "",
            "club": "",
            "chip": "c3",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(status=ResultStatus.OK, time=13),
            "start": start_type.PersonRaceStart(),
        },
    ]


def test_do_not_import_special_names():
    value_1 = "c1,,Vacant,,,0:11,,,,0,,"
    value_2 = "c2,v2,n2,,,0:12,,,,0,,"
    value_3 = "c3,,Reserve,,,0:13,,,,0,,"
    content = bytes(
        ",".join(header) + "\n" + value_1 + "\n" + value_2 + "\n" + value_3,
        encoding="utf-8",
    )
    assert parse(content) == [
        {
            "first_name": "v2",
            "last_name": "n2",
            "class_": "",
            "club": "",
            "chip": "c2",
            "gender": "",
            "year": None,
            "not_competing": False,
            "result": result_type.PersonRaceResult(status=ResultStatus.OK, time=12),
            "start": start_type.PersonRaceStart(),
        }
    ]


def test_import_extra_fields():
    value_1 = "c1,v1,n1,,,0:11,,,,0,,,A,B,C"
    value_2 = "c2,v2,n2,,,0:12,,,,0,,,,X,"
    value_3 = "c3,v3,n3,,,0:13,,,,0,,,,Y,Z"
    content = bytes(
        ",".join(header + ["Text1", "Text2", "Text3"])
        + "\n"
        + value_1
        + "\n"
        + value_2
        + "\n"
        + value_3,
        encoding="utf-8",
    )
    assert parse(content) == [
        {
            "first_name": "v1",
            "last_name": "n1",
            "class_": "",
            "club": "",
            "chip": "c1",
            "gender": "",
            "year": None,
            "not_competing": False,
            "fields": {0: "A", 1: "B", 2: "C"},
            "result": result_type.PersonRaceResult(status=ResultStatus.OK, time=11),
            "start": start_type.PersonRaceStart(),
        },
        {
            "first_name": "v2",
            "last_name": "n2",
            "class_": "",
            "club": "",
            "chip": "c2",
            "gender": "",
            "year": None,
            "not_competing": False,
            "fields": {0: "", 1: "X", 2: ""},
            "result": result_type.PersonRaceResult(status=ResultStatus.OK, time=12),
            "start": start_type.PersonRaceStart(),
        },
        {
            "first_name": "v3",
            "last_name": "n3",
            "class_": "",
            "club": "",
            "chip": "c3",
            "gender": "",
            "year": None,
            "not_competing": False,
            "fields": {0: "", 1: "Y", 2: "Z"},
            "result": result_type.PersonRaceResult(status=ResultStatus.OK, time=13),
            "start": start_type.PersonRaceStart(),
        },
    ]



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import dataclasses
from typing import Optional


@dataclasses.dataclass
class Settings:
    name: str = ""
    nr_of_best_results: Optional[int] = None
    mode: str = "Proportional"
    maximum_points: int = 100
    decimal_places: int = 2

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def get(self, key, default):
        return getattr(self, key, default)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pathlib
from datetime import datetime

from fpdf import FPDF

from ooresults.repo.result_type import ResultStatus


class PDF(FPDF):
    MAP_STATUS = {
        ResultStatus.INACTIVE: "",
        ResultStatus.FINISHED: "Finished",
        ResultStatus.OK: "OK",
        ResultStatus.MISSING_PUNCH: "MP",
        ResultStatus.DID_NOT_START: "DNS",
        ResultStatus.DID_NOT_FINISH: "DNF",
        ResultStatus.OVER_TIME: "OTL",
        ResultStatus.DISQUALIFIED: "DSQ",
    }

    def __init__(self, name: str, landscape: bool = False):
        orientation = "landscape" if landscape else "portrait"
        super().__init__(font_cache_dir=None, orientation=orientation)
        self.name = name
        self.creation_time = datetime.now()

        self.fonts_dir = pathlib.Path(__file__).parent / "fonts"
        self.add_font(
            family="Carlito",
            fname=str(self.fonts_dir / "Carlito-Regular.ttf"),
            uni=True,
        )
        self.add_font(
            family="Carlito",
            style="B",
            fname=str(self.fonts_dir / "Carlito-Bold.ttf"),
            uni=True,
        )
        self.add_font(
            family="Carlito",
            style="I",
            fname=str(self.fonts_dir / "Carlito-Italic.ttf"),
            uni=True,
        )
        self.add_font(
            family="Carlito",
            style="BI",
            fname=str(self.fonts_dir / "Carlito-BoldItalic.ttf"),
            uni=True,
        )

    def header(self):
        self.set_font(family="Carlito", size=12)
        self.cell(w=0, h=10, txt=self.name, border=1, align="C")
        self.ln(20)

    def footer(self):
        # Position cursor at 1.5 cm from bottom:
        self.set_y(-15)
        self.set_font(family="Carlito", size=12)
        # Printing page number
        self.cell(
            w=0, h=10, txt=self.creation_time.strftime("%Y-%m-%d %H:%M:%S"), align="L"
        )
        self.cell(w=0, h=10, txt=f"Page {self.page_no()}/{{nb}}", align="R")



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import pathlib

import web
from web.utils import Storage

from ooresults.handler import model
from ooresults.repo.repo import CourseUsedError
from ooresults.repo.repo import EventNotFoundError
from ooresults.repo.repo import ConstraintError
from ooresults.plugins import iof_course_data


templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals={"str": str, "round": round})


def update(event_id: int):
    courses_list = model.get_courses(event_id=event_id)
    event = list(model.get_event(event_id))
    if event == []:
        return render.courses_table({}, courses_list)
    else:
        return render.courses_table(event[0], courses_list)


class Update:
    def POST(self):
        """Update data"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        return update(event_id)


class Import:
    def POST(self):
        """Import entries"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            event = {}
            if data.cou_import == "cou.import.1":
                event, courses, class_course = iof_course_data.parse_course_data(
                    data.browse1
                )
                model.import_courses(
                    event_id=event_id, courses=courses, class_course=class_course
                )

        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except Exception as e:
            raise web.Conflict(str(e))

        return update(event_id)


class Export:
    def POST(self):
        """Export entries"""
        data = web.input()
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            if data.cou_export == "cou.export.1":
                event = model.get_event(id=event_id)
                courses = model.get_courses(event_id=event_id)
                classes = model.get_classes(event_id=event_id)
                content = iof_course_data.create_course_data(event[0], courses, classes)

        except KeyError:
            raise web.conflict("Entry deleted")
        except:
            logging.exception("Internal server error")
            raise

        return content


class Add:
    def POST(self):
        """Add or edit entry"""
        data = web.input()
        print(data)
        event_id = int(data.event_id) if data.event_id != "" else -1
        length = int(data.length) if data.length != "" else None
        climb = int(data.climb) if data.climb != "" else None
        controls = data.controls.split("-") if data.controls != "" else []
        controls = [c.strip() for c in controls]
        try:
            if data.id == "":
                model.add_course(
                    event_id=event_id,
                    name=data.name,
                    length=length,
                    climb=climb,
                    controls=controls,
                )
            else:
                model.update_course(
                    id=int(data.id),
                    event_id=event_id,
                    name=data.name,
                    length=length,
                    climb=climb,
                    controls=controls,
                )

        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except ConstraintError as e:
            raise web.conflict(str(e))
        except KeyError:
            raise web.conflict("Course deleted")
        except:
            logging.exception("Internal server error")
            raise

        return update(event_id)


class Delete:
    def POST(self):
        """Delete entry"""
        data = web.input()
        print(data)
        event_id = int(data.event_id) if data.event_id != "" else -1
        try:
            model.delete_course(id=int(data.id))
            return update(event_id)
        except CourseUsedError:
            raise web.conflict("Course used in classes")
        except:
            logging.exception("Internal server error")
            raise


class FillEditForm:
    def POST(self):
        """Query data to fill add or edit form"""
        data = web.input()
        try:
            if data.id == "":
                course = Storage(
                    {
                        "id": "",
                        "name": "",
                        "length": None,
                        "climb": None,
                        "controls": "",
                    }
                )
            else:
                course = model.get_course(id=int(data.id))[0]
        except EventNotFoundError:
            raise web.conflict("No event selected or event deleted")
        except KeyError:
            raise web.conflict("Course deleted")
        except:
            logging.exception("Internal server error")
            raise

        return render.add_course(course)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

from ooresults.plugins import iof_course_data


def test_import_course_data():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<CourseData xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <RaceCourseData>
    <Course>
      <Name>Bahn A</Name>
      <Length>4800</Length>
      <Climb>120</Climb>
      <CourseControl type="Start">
        <Control>S1</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>101</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>102</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>103</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>104</Control>
      </CourseControl>
      <CourseControl type="Finish">
        <Control>F1</Control>
      </CourseControl>
    </Course>
    <Course>
      <Name>Bahn B</Name>
      <CourseControl type="Start">
        <Control>S1</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>101</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>105</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>106</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>104</Control>
      </CourseControl>
      <CourseControl type="Finish">
        <Control>F1</Control>
      </CourseControl>
    </Course>
  </RaceCourseData>
</CourseData>
"""
    event, courses, class_course = iof_course_data.parse_course_data(
        bytes(content, encoding="utf-8")
    )
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert courses == [
        {
            "name": "Bahn A",
            "length": 4800,
            "climb": 120,
            "controls": ["101", "102", "103", "104"],
        },
        {
            "name": "Bahn B",
            "length": None,
            "climb": None,
            "controls": ["101", "105", "106", "104"],
        },
    ]
    assert class_course == []


def test_import_course_data_with_class_assignment():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<CourseData xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <RaceCourseData>
    <Course>
      <Name>Bahn A</Name>
      <Length>4800</Length>
      <Climb>120</Climb>
      <CourseControl type="Start">
        <Control>S1</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>101</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>102</Control>
      </CourseControl>
      <CourseControl type="Finish">
        <Control>F1</Control>
      </CourseControl>
    </Course>
    <Course>
      <Name>Bahn B</Name>
      <CourseControl type="Start">
        <Control>S1</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>104</Control>
      </CourseControl>
      <CourseControl type="Finish">
        <Control>F1</Control>
      </CourseControl>
    </Course>
    <ClassCourseAssignment>
      <ClassName>Elite Men</ClassName>
      <CourseName>Bahn B</CourseName>
    </ClassCourseAssignment>
    <ClassCourseAssignment>
      <ClassName>Elite Woman</ClassName>
      <CourseName>Bahn A</CourseName>
    </ClassCourseAssignment>
    <ClassCourseAssignment>
      <ClassName>Beginners</ClassName>
    </ClassCourseAssignment>
  </RaceCourseData>
</CourseData>
"""
    event, courses, class_course = iof_course_data.parse_course_data(
        bytes(content, encoding="utf-8")
    )
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert courses == [
        {
            "name": "Bahn A",
            "length": 4800,
            "climb": 120,
            "controls": ["101", "102"],
        },
        {
            "name": "Bahn B",
            "length": None,
            "climb": None,
            "controls": ["104"],
        },
    ]
    assert class_course == [
        {
            "class_name": "Elite Men",
            "course_name": "Bahn B",
        },
        {
            "class_name": "Elite Woman",
            "course_name": "Bahn A",
        },
        {
            "class_name": "Beginners",
            "course_name": None,
        },
    ]


def test_export_course_data():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<CourseData xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <RaceCourseData>
    <Course>
      <Name>Bahn A</Name>
      <Length>4800</Length>
      <Climb>120</Climb>
      <CourseControl type="Start">
        <Control>S1</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>101</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>102</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>103</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>104</Control>
      </CourseControl>
      <CourseControl type="Finish">
        <Control>F1</Control>
      </CourseControl>
    </Course>
    <Course>
      <Name>Bahn B</Name>
      <CourseControl type="Start">
        <Control>S1</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>101</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>105</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>106</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>104</Control>
      </CourseControl>
      <CourseControl type="Finish">
        <Control>F1</Control>
      </CourseControl>
    </Course>
  </RaceCourseData>
</CourseData>
"""
    document = iof_course_data.create_course_data(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [
            {
                "name": "Bahn A",
                "length": 4800,
                "climb": 120,
                "controls": ["101", "102", "103", "104"],
            },
            {
                "name": "Bahn B",
                "length": None,
                "climb": None,
                "controls": ["101", "105", "106", "104"],
            },
        ],
        [],
    )
    assert document == bytes(content, encoding="utf-8")


def test_export_course_data_with_class_assignment():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<CourseData xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <RaceCourseData>
    <Course>
      <Name>Bahn A</Name>
      <Length>4800</Length>
      <Climb>120</Climb>
      <CourseControl type="Start">
        <Control>S1</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>101</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>102</Control>
      </CourseControl>
      <CourseControl type="Finish">
        <Control>F1</Control>
      </CourseControl>
    </Course>
    <Course>
      <Name>Bahn B</Name>
      <CourseControl type="Start">
        <Control>S1</Control>
      </CourseControl>
      <CourseControl type="Control">
        <Control>104</Control>
      </CourseControl>
      <CourseControl type="Finish">
        <Control>F1</Control>
      </CourseControl>
    </Course>
    <ClassCourseAssignment>
      <ClassName>Elite Men</ClassName>
      <CourseName>Bahn B</CourseName>
    </ClassCourseAssignment>
    <ClassCourseAssignment>
      <ClassName>Elite Woman</ClassName>
      <CourseName>Bahn A</CourseName>
    </ClassCourseAssignment>
    <ClassCourseAssignment>
      <ClassName>Beginners</ClassName>
    </ClassCourseAssignment>
  </RaceCourseData>
</CourseData>
"""
    document = iof_course_data.create_course_data(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [
            {
                "name": "Bahn A",
                "length": 4800,
                "climb": 120,
                "controls": ["101", "102"],
            },
            {
                "name": "Bahn B",
                "length": None,
                "climb": None,
                "controls": ["104"],
            },
        ],
        [
            {
                "name": "Elite Men",
                "course": "Bahn B",
            },
            {
                "name": "Elite Woman",
                "course": "Bahn A",
            },
            {
                "name": "Beginners",
                "course": None,
            },
        ],
    )
    assert document == bytes(content, encoding="utf-8")



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pathlib
import tempfile
from unittest.mock import patch

import pytest

from ooresults import configuration


def test_configuration_create_default_config_if_not_exists():
    with tempfile.TemporaryDirectory() as hd:
        home = pathlib.Path(hd)

        def my_home() -> pathlib.Path:
            return home

        with patch.object(pathlib.Path, "home", my_home):
            c = configuration.Config(path=home / ".ooresults")
            assert c.ssl_cert == home / ".ooresults" / "cert" / "cert.pem"
            assert c.ssl_key == home / ".ooresults" / "cert" / "privkey.pem"
            assert c.demo_reader is False

            assert c.ssl_cert.exists()
            assert c.ssl_key.exists()
            assert (home / ".ooresults" / "config.ini").exists()


def test_configuration_cert_files_are_created_in_home():
    with tempfile.TemporaryDirectory() as td:
        with tempfile.TemporaryDirectory() as hd:
            temp = pathlib.Path(td)
            home = pathlib.Path(hd)

            def my_home() -> pathlib.Path:
                return home

            with patch.object(pathlib.Path, "home", my_home):
                c = configuration.Config(path=temp)
                assert c.ssl_cert == home / ".ooresults" / "cert" / "cert.pem"
                assert c.ssl_key == home / ".ooresults" / "cert" / "privkey.pem"
                assert c.demo_reader is False

                assert c.ssl_cert.exists()
                assert c.ssl_key.exists()
                assert c.ssl_cert.stat().st_size > 100
                assert c.ssl_key.stat().st_size > 100
                assert (temp / "config.ini").exists()


def test_configuration_cert_files_are_only_created_if_not_exist():
    with tempfile.TemporaryDirectory() as td:
        with tempfile.TemporaryDirectory() as hd:
            temp = pathlib.Path(td)
            home = pathlib.Path(hd)

            def my_home() -> pathlib.Path:
                return home

            ssl_cert = home / ".ooresults" / "cert" / "cert.pem"
            ssl_key = home / ".ooresults" / "cert" / "privkey.pem"
            ssl_cert.parent.mkdir(parents=True, exist_ok=True)
            ssl_cert.touch()
            ssl_key.touch()

            with patch.object(pathlib.Path, "home", my_home):
                c = configuration.Config(path=temp)
                assert ssl_cert.stat().st_size == 0
                assert ssl_key.stat().st_size == 0
                assert c.ssl_cert == ssl_cert
                assert c.ssl_key == ssl_key


def test_configuration_cert_files_are_not_created_if_defined_in_config():
    with tempfile.TemporaryDirectory() as td:
        with tempfile.TemporaryDirectory() as hd:
            temp = pathlib.Path(td)
            home = pathlib.Path(hd)

            def my_home() -> pathlib.Path:
                return home

            ssl_cert = temp / "cert.pem"
            ssl_key = temp / "privkey.pem"
            ssl_cert.touch()
            ssl_key.touch()

            config_file = temp / "config.ini"
            with open(config_file, "w") as f:
                f.write("[Server]\n")
                f.write(f"ssl_cert = {str(ssl_cert)}\n")
                f.write(f"ssl_key = {str(ssl_key)}\n")

            with patch.object(pathlib.Path, "home", my_home):
                c = configuration.Config(path=temp)

                assert not (home / ".ooresults" / "cert" / "cert.pem").exists()
                assert not (home / ".ooresults" / "cert" / "privkey.pem").exists()
                assert not (home / "config.ini").exists()

                assert c.ssl_cert == ssl_cert
                assert c.ssl_key == ssl_key
                assert c.ssl_cert.stat().st_size == 0
                assert c.ssl_key.stat().st_size == 0


def test_configuration_exception_if_cert_files_defined_but_not_found():
    with tempfile.TemporaryDirectory() as td:
        with tempfile.TemporaryDirectory() as hd:
            temp = pathlib.Path(td)
            home = pathlib.Path(hd)

            def my_home() -> pathlib.Path:
                return home

            ssl_cert = temp / "cert.pem"
            ssl_key = temp / "privkey.pem"

            config_file = temp / "config.ini"
            with open(config_file, "w") as f:
                f.write("[Server]\n")
                f.write(f"ssl_cert = {str(ssl_cert)}\n")
                f.write(f"ssl_key = {str(ssl_key)}\n")

            with patch.object(pathlib.Path, "home", my_home):
                with pytest.raises(
                    expected_exception=FileNotFoundError,
                    match=f"Certificate file '{ssl_cert}' not found",
                ):
                    c = configuration.Config(path=temp)

                assert not (home / ".ooresults" / "cert" / "cert.pem").exists()
                assert not (home / ".ooresults" / "cert" / "privkey.pem").exists()
                assert not (home / "config.ini").exists()


def test_configuration_demo_reader_default_is_false():
    with tempfile.TemporaryDirectory() as td:
        home = pathlib.Path(td)

        def my_home() -> pathlib.Path:
            return home

        with patch.object(pathlib.Path, "home", my_home):
            config_file = home / "config.ini"
            with open(config_file, "w") as f:
                f.write("[Server]\n")

            c = configuration.Config(path=home)
            assert c.demo_reader is False


def test_configuration_config_file_is_read_if_exists_1():
    with tempfile.TemporaryDirectory() as td:
        home = pathlib.Path(td)

        def my_home() -> pathlib.Path:
            return home

        with patch.object(pathlib.Path, "home", my_home):
            config_file = home / "config.ini"
            with open(config_file, "w") as f:
                f.write("[Server]\n")
                f.write("demo_reader = off")

            c = configuration.Config(path=home)
            assert c.demo_reader is False


def test_configuration_config_file_is_read_if_exists_2():
    with tempfile.TemporaryDirectory() as td:
        home = pathlib.Path(td)

        def my_home() -> pathlib.Path:
            return home

        with patch.object(pathlib.Path, "home", my_home):
            config_file = home / "config.ini"
            with open(config_file, "w") as f:
                f.write("[Server]\n")
                f.write("demo_reader = on")

            c = configuration.Config(path=home)
            assert c.demo_reader is True



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging

from ooresults.repo.update import update_008


VERSION = 8


def update_tables(db, path: str = "ooresults.sqlite") -> None:
    t = db.transaction()
    db.ctx.db.execute("BEGIN EXCLUSIVE TRANSACTION;")

    values = list(db.query("SELECT value FROM version"))
    if not values:
        raise RuntimeError("DB error - table version is empty")

    else:
        version = values[0].value
        logging.info(f"DB version is {version}")

        if version > VERSION:
            raise RuntimeError(
                f"DB version to high - version = {version}, but must be at most {VERSION}"
            )

        elif version < VERSION:
            t.rollback()

            if version <= 6:
                raise RuntimeError(
                    f"DB version to low - version = {version}, but must be at least 7"
                )
            if version <= 7:
                logging.info("Update DB to version 8 ...")
                update_008.update(path=path)

            logging.info(f"DB updated to version {VERSION}")
        else:
            t.rollback()



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from tests.competitor import Competitor
from ooresults.plugins import iof_competitor_list


def test_import_name():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<CompetitorList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Competitor>
    <Person>
      <Name>
        <Family>Merkel</Family>
        <Given>Angela</Given>
      </Name>
    </Person>
  </Competitor>
</CompetitorList>
"""
    assert iof_competitor_list.parse_competitor_list(
        bytes(content, encoding="utf-8")
    ) == [
        {
            "first_name": "Angela",
            "last_name": "Merkel",
            "club": "",
            "chip": "",
            "gender": "",
            "year": None,
        }
    ]


def test_export_name():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<CompetitorList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Competitor>
    <Person>
      <Name>
        <Family>Merkel</Family>
        <Given>Angela</Given>
      </Name>
    </Person>
  </Competitor>
</CompetitorList>
"""
    document = iof_competitor_list.create_competitor_list(
        [
            Competitor(first_name="Angela", last_name="Merkel"),
        ]
    )
    assert document == bytes(content, encoding="utf-8")


def test_import_full():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<CompetitorList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Competitor>
    <Person sex="F">
      <Name>
        <Family>Merkel</Family>
        <Given>Angela</Given>
      </Name>
      <BirthDate>1972-01-01</BirthDate>
    </Person>
    <Organisation>
      <Name>OC Kanzleramt</Name>
    </Organisation>
    <ControlCard punchingSystem="SI">1234567</ControlCard>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
  </Competitor>
</CompetitorList>
"""
    assert iof_competitor_list.parse_competitor_list(
        bytes(content, encoding="utf-8")
    ) == [
        {
            "first_name": "Angela",
            "last_name": "Merkel",
            "club": "OC Kanzleramt",
            "chip": "1234567",
            "gender": "F",
            "year": 1972,
        }
    ]


def test_export_full():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<CompetitorList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Competitor>
    <Person sex="F">
      <Name>
        <Family>Merkel</Family>
        <Given>Angela</Given>
      </Name>
      <BirthDate>1972-01-01</BirthDate>
    </Person>
    <Organisation>
      <Name>OC Kanzleramt</Name>
    </Organisation>
    <ControlCard punchingSystem="SI">1234567</ControlCard>
  </Competitor>
</CompetitorList>
"""
    document = iof_competitor_list.create_competitor_list(
        [
            Competitor(
                first_name="Angela",
                last_name="Merkel",
                club_name="OC Kanzleramt",
                gender="F",
                year=1972,
                chip="1234567",
            )
        ]
    )
    assert document == bytes(content, encoding="utf-8")
